#include <math.h>
#include "main.h"
#include "game.h"
#include "screen.h"
#include "drawpod.h"
#include "drawblt.h"
#include "explode.h"
#include "timeloop.h"
#include "drawpwrp.h"
#include "fractal.h"



static void draw_world(BITMAP *bmp, GAME *game, POD *camera_pod)
{
	clear(bmp);

	/* Evil polygon routines */
	bmp->cl++; bmp->cr--; bmp->ct++; bmp->cb--;

	color_map = &the_trans_map;

	{
		float phi = (float)camera_pod->phi * (3.14159265358979323846f / (128 << 16));
		float sp = sin(phi), cp = cos(phi);
		float st = sin(-0.5f), ct = cos(-0.5f);

		float camera_forward = (camera_pod->flags & POD_EXTERNAL_VIEW) || camera_pod->state == POD_STATE_DEAD ? EXT_CAMERA_FORWARD : CAMERA_FORWARD;
		float camera_height = (camera_pod->flags & POD_EXTERNAL_VIEW) || camera_pod->state == POD_STATE_DEAD ? EXT_CAMERA_HEIGHT : CAMERA_HEIGHT;

		float xc = fixtof(camera_pod->x) + sp*camera_forward;
		float yc = fixtof(camera_pod->y) - cp*camera_forward;
		float zc = fixtof(camera_pod->z) + camera_height;

		MATRIX_f camera = {
			{
				{cp               ,  sp               ,  0             },
				{sp*st*(4.0f/3.0f), -cp*st*(4.0f/3.0f), -ct*(4.0f/3.0f)},
				{sp*ct            , -cp*ct            , st             }
			},
			{
				-(xc*cp + yc*sp),
				-(xc*sp*st - yc*cp*st - zc*ct) * (4.0f/3.0f),
				-(xc*sp*ct - yc*cp*ct + zc*st)
			}
		};

		{
			int i;
			for (i = 0; i < game->n_vertices; i++) draw_prepare_vertex(&camera, &game->vertex[i]);
		}
		{
			POD *pod;
			for (pod = game->pod; pod; pod = pod->next) prepare_pod_vertices(&camera, pod);
		}
		{
			BULLET *bullet;
			for (bullet = game->bullet; bullet; bullet = bullet->next) prepare_bullet_vertices(&camera, bullet);
		}
		{
			EXPLOSION *explosion;
			for (explosion = game->explosion; explosion; explosion = explosion->next) prepare_explosion_vertices(&camera, explosion);
		}
		{
			POWERUP *powerup;
			for (powerup = game->powerup; powerup; powerup = powerup->next) prepare_powerup_vertices(&camera, powerup);
		}
		draw_prepare_bsp_tree(&camera, game->bsp_tree);
		set_projection_viewport(0, 0, bmp->w, bmp->h);
		draw_bsp_tree(bmp, &camera, xc, yc, zc, game->bsp_tree);
	}

	/* Evil polygon routines */
	bmp->cl--; bmp->cr++; bmp->ct--; bmp->cb++;
	rect(bmp, bmp->cl, bmp->ct, bmp->cr-1, bmp->cb-1, 0);
}



static void draw_view(BITMAP *bmp, GAME *game, POD *pod)
{
	draw_world(bmp, game, pod);

	color_map = &the_trans_map;
	drawing_mode(DRAW_MODE_TRANS, NULL, 0, 0);
	{
		int xl = bmp->w >> 6;
		int xr = bmp->w - xl;
		int yt = bmp->h >> 6;
		int yb = bmp->h*3 >> 6;
		int h = pod->health & (POD_HEALTH_INITIAL - 1);
		int x = xl + (h * (xr - xl) >> POD_HEALTH_INITIAL_SHIFT);
		int c1 = pod->health >= POD_HEALTH_INITIAL ? COLOUR_YELLOW : COLOUR_RED;
		int c2 = pod->health >= POD_HEALTH_INITIAL ? COLOUR_RED : COLOUR_DARK_GREY;
		if (x > xl) rectfill(bmp, xl, yt, x-1, yb-1, c1);
		if (x < xr) rectfill(bmp, x, yt, xr-1, yb-1, c2);
	}
	solid_mode();
}



void draw_game(void *data)
{
	GAME *game = data;
	BITMAP *bmp = get_screen();

	render_fractal();

	clear(bmp);

	{
		POD *pod;
		for (pod = game->pod; pod; pod = pod->next) add_pod_polygons(game->bsp_tree, pod);
	}
	{
		BULLET *bullet;
		for (bullet = game->bullet; bullet; bullet = bullet->next) add_bullet_polygons(game->bsp_tree, bullet);
	}
	{
		EXPLOSION *explosion;
		for (explosion = game->explosion; explosion; explosion = explosion->next) add_explosion_polygons(game->bsp_tree, explosion);
	}
	{
		POWERUP *powerup;
		for (powerup = game->powerup; powerup; powerup = powerup->next) add_powerup_polygons(game->bsp_tree, powerup);
	}
	finalise_bsp_tree(game->bsp_tree, 1);
	{
		POD *pod;
		for (pod = game->pod; pod; pod = pod->next) {
			if (pod->view_dim) {
				BITMAP *sbmp = create_sub_bitmap(bmp,
					bmp->w * pod->view_dim[0] >> 2,
					bmp->h * pod->view_dim[1] >> 2,
					bmp->w * pod->view_dim[2] >> 2,
					bmp->h * pod->view_dim[3] >> 2);
				if (sbmp) {
					draw_view(sbmp, game, pod);
					destroy_bitmap(sbmp);
				}
			}
		}
	}
	remove_temporary(game->bsp_tree);

	if (show_fps) {
		text_mode(-1);
		textprintf_right(bmp, font, bmp->w - 2, bmp->h - 10, (int)FLOAT_COLOUR_EXPLOSION_BRIGHT, "FPS:%d Speed:%d", fps_n_draws, fps_n_updates);
	}

	commit_screen();
}
