#include <stdlib.h>
#include <time.h>
#include "main.h"
#include "screen.h"
#include "game.h"
#include "timeloop.h"



DATAFILE *dat;
COLOR_MAP the_trans_map;
BITMAP *the_fractal_bitmap;
int joystick_installed;
int show_fps;



void poll_stuff(void)
{
	poll_the_controls();
}



static void add_blend(const PALETTE pal, int x, int y, RGB *rgb)
{
	rgb->r = MIN(pal[x].r + pal[y].r, 63);
	rgb->g = MIN(pal[x].g + pal[y].g, 63);
	rgb->b = MIN(pal[x].b + pal[y].b, 63);
}



static void calibrate_joysticks(void)
{
	int i;
	for (i = 0; i < num_joysticks; i++) {
		while (joy[i].flags & JOYFLAG_CALIBRATE) {
			clear(screen);
			textprintf_centre(screen, font, SCREEN_W/2, SCREEN_H/2-9, (int)FLOAT_COLOUR_EXPLOSION_BRIGHT, "Joystick %d", i + 1);
			textout_centre(screen, font, calibrate_joystick_name(i), SCREEN_W/2, SCREEN_H/2+1, COLOUR_YELLOW);
			calibrate_joystick(i);
		}
	}
}



static void check_joy_descriptor(JOY_DESCRIPTOR *jd)
{
	if (jd->joystick != 255) {
		if (jd->joystick < num_joysticks) {
			JOYSTICK_INFO *j = &joy[jd->joystick];
			if (jd->stick < j->num_sticks) {
				JOYSTICK_STICK_INFO *js = &j->stick[jd->stick];
				if ((jd->axis & 0x7F) < js->num_axis /* axes, grr */ ) {
					/* Valid */
				} else
					jd->joystick = 255;
			} else if (jd->stick == 255) {
				if (jd->axis < j->num_buttons) {
					/* Valid */
				} else
					jd->joystick = 255;
			} else
				jd->joystick = 255;
		} else
			jd->joystick = 255;
	}
}



static void validate_joystick_controls(USER_KEYS *k)
{
	calibrate_joysticks();
	check_joy_descriptor(&k->jleft);
	check_joy_descriptor(&k->jright);
	check_joy_descriptor(&k->jaccel);
	check_joy_descriptor(&k->jland);
	check_joy_descriptor(&k->jfire);
}



static unsigned char validate_key(unsigned char key)
{
	if (key == KEY_ESC || key == KEY_PAUSE || key >= KEY_MAX) return 0;
	return key;
}



int main(void);
int main(void)
{
	srand(time(NULL));
	if (allegro_init()) return 1;
	if (install_keyboard()) return 1;
	if (install_timer()) return 1;
	init_timeloop();
	dat = load_datafile("podfight.dat");
	if (!dat) {
		allegro_message("Unable to load the datafile!\n");
		return 1;
	}
	the_fractal_bitmap = create_bitmap(256, 256);
	if (!the_fractal_bitmap) {
		allegro_message("Out of memory!\n");
		return 1;
	}
	if (set_screen(8, get_config_int("podfight", "gfx_w", 640), get_config_int("podfight", "gfx_h", 480)) != 0) {
		allegro_message("Unable to set graphics mode!\n");
		return 1;
	}
	create_color_table(&the_trans_map, dat[THE_PALETTE].dat, &add_blend, NULL);

	joystick_installed = install_joystick(JOY_TYPE_AUTODETECT) == 0;

	show_fps = get_config_int("podfight", "show_fps", 0);

	{
		USER_KEYS uk[4];
		GAME *game;
		int n_players = get_config_int("podfight", "n_players", 1);
		int i;
		n_players = MID(1, n_players, 4);
		for (i = 0; i < n_players; i++) {
			static char section_name[] = "podfight_player0";
			section_name[15] = '1' + i;
			uk[i].device = joystick_installed && get_config_int(section_name, "use_joystick", 0) ? UK_DEVICE_JOYSTICK : UK_DEVICE_KEYBOARD;
			uk[i].left  = validate_key(get_config_int(section_name, "key_left" , 0));
			uk[i].right = validate_key(get_config_int(section_name, "key_right", 0));
			uk[i].accel = validate_key(get_config_int(section_name, "key_accel", 0));
			uk[i].land  = validate_key(get_config_int(section_name, "key_land" , 0));
			uk[i].fire  = validate_key(get_config_int(section_name, "key_fire" , 0));
			uk[i].jleft .joystick =  get_config_int(section_name, "joy_left_joystick" , 0) - 1;
			uk[i].jleft .stick    =  get_config_int(section_name, "joy_left_stick"    , 1) - 1;
			uk[i].jleft .axis     = (get_config_int(section_name, "joy_left_axis"     , 1) - 1) | (uk[i].jleft .stick != 255 && get_config_int(section_name, "joy_left_sign" , -1) < 0 ? 128 : 0);
			uk[i].jright.joystick =  get_config_int(section_name, "joy_right_joystick", 0) - 1;
			uk[i].jright.stick    =  get_config_int(section_name, "joy_right_stick"   , 1) - 1;
			uk[i].jright.axis     = (get_config_int(section_name, "joy_right_axis"    , 1) - 1) | (uk[i].jright.stick != 255 && get_config_int(section_name, "joy_right_sign", +1) < 0 ? 128 : 0);
			uk[i].jaccel.joystick =  get_config_int(section_name, "joy_accel_joystick", 0) - 1;
			uk[i].jaccel.stick    =  get_config_int(section_name, "joy_accel_stick"   , 1) - 1;
			uk[i].jaccel.axis     = (get_config_int(section_name, "joy_accel_axis"    , 2) - 1) | (uk[i].jaccel.stick != 255 && get_config_int(section_name, "joy_accel_sign", -1) < 0 ? 128 : 0);
			uk[i].jland .joystick =  get_config_int(section_name, "joy_land_joystick" , 0) - 1;
			uk[i].jland .stick    =  get_config_int(section_name, "joy_land_stick"    , 1) - 1;
			uk[i].jland .axis     = (get_config_int(section_name, "joy_land_axis"     , 2) - 1) | (uk[i].jland .stick != 255 && get_config_int(section_name, "joy_land_sign" , +1) < 0 ? 128 : 0);
			uk[i].jfire .joystick =  get_config_int(section_name, "joy_fire_joystick" , 0) - 1;
			uk[i].jfire .stick    =  get_config_int(section_name, "joy_fire_stick"    , 0) - 1;
			uk[i].jfire .axis     = (get_config_int(section_name, "joy_fire_axis"     , 1) - 1) | (uk[i].jfire .stick != 255 && get_config_int(section_name, "joy_fire_sign" ,  0) < 0 ? 128 : 0);
			validate_joystick_controls(&uk[i]);
		}
		game = create_game(n_players);
		if (!game) return 1;
		for (i = 0; i < n_players; i++) timeloop_add_keys(&uk[i], i);
		timeloop(BPS_TO_TIMER(100), 20, &draw_game, &update_game, game);
		for (i = 0; i < n_players; i++) timeloop_remove_keys(&uk[i], i);
		destroy_game(game);
	}

	unload_datafile(dat);
	return 0;
}
END_OF_MAIN()
