#include <string.h>
#include "main.h"
#include "timeloop.h"
#include "controls.h"



volatile int true_time;
int logic_time;

static volatile int max_true_time;

static int timeloop_active = 0;

static int timeloop_speed;



/* (float)fps_n_draws * BPS / FPS_FRAMES gives the frames per second.
 * (float)fps_n_updates / FPS_FRAMES gives 1.0 if the game is running up to
 * speed, and it may drop if the FPS drops too low.
 */
int fps_n_updates;
int fps_n_draws;

static volatile int n_skipped;
static int n_skipped_base;

#define FPS_UPDATE_CALLED 1
#define FPS_DRAW_CALLED   2

static unsigned char fps_flags[FPS_FRAMES];
static unsigned char fps_flags_current;
static int fps_ptr = 0;

#define FPS_FLAGS_SET(flags)                                        \
{                                                                   \
	fps_n_updates += ((flags & 1) - (fps_flags[fps_ptr] & 1));      \
	fps_n_draws   += ((flags & 2) - (fps_flags[fps_ptr] & 2)) >> 1; \
	fps_flags[fps_ptr] = flags;                                     \
	fps_ptr = (fps_ptr + 1) % FPS_FRAMES;                           \
}



#define MAX_SKIP 15 /* Must be 2^n-1 for some integer n. */
#define N_KEY_SETS 4

static USER_KEYS *timeloop_keys[N_KEY_SETS] = {NULL, NULL};
static volatile USER_KEYS_HELD timeloop_keys_held[MAX_SKIP+1][N_KEY_SETS];
static volatile int timeloop_true_ptr;
static int timeloop_logic_ptr;
static volatile int need_poll;
USER_KEYS_HELD *ckey;

/* These are all signed. */
static volatile int joystick_left[N_KEY_SETS];
static volatile int joystick_right[N_KEY_SETS];
static volatile int joystick_accel[N_KEY_SETS];
static volatile int joystick_land[N_KEY_SETS];
static volatile int joystick_fire[N_KEY_SETS];



int get_joystick_state(JOY_DESCRIPTOR *jd)
{
	if (jd->joystick < num_joysticks) {
		JOYSTICK_INFO *j = &joy[jd->joystick];
		if (jd->stick < j->num_sticks) {
			JOYSTICK_STICK_INFO *js = &j->stick[jd->stick];
			if ((jd->axis & 0x7F) < js->num_axis /* axes, grr */ ) {
				JOYSTICK_AXIS_INFO *ja = &js->axis[jd->axis & 0x7F];
				int rv = ja->pos;
				if (js->flags & JOYFLAG_UNSIGNED) rv -= 128;
				if (jd->axis & 0x80) rv = -rv;
				return rv;
			}
		} else if (jd->axis < j->num_buttons) {
			JOYSTICK_BUTTON_INFO *jb = &j->button[jd->axis];
			return jb->b ? 128 : 0;
		}
	}
	return 0;
}



int always_poll_joystick = 0;



void poll_the_controls(void)
{
	if (need_poll) {
		int i;
		int polled = always_poll_joystick;
		if (always_poll_joystick) poll_joystick();
		for (i = 0; i < N_KEY_SETS; i++) {
			USER_KEYS *k = timeloop_keys[i];
			if (k && k->device == UK_DEVICE_JOYSTICK) {
				/* We don't call poll_joystick() unless the joysticks are
				 * actually in use. This innocent-looking function can slow
				 * execution down considerably.
				 */
				if (!polled) {
					poll_joystick();
					polled = 1;
				}
				joystick_left [i] = get_joystick_state(&k->jleft);
				joystick_right[i] = get_joystick_state(&k->jright);
				joystick_accel[i] = get_joystick_state(&k->jaccel);
				joystick_land [i] = get_joystick_state(&k->jland);
				joystick_fire [i] = get_joystick_state(&k->jfire);
			}
		}
		need_poll = 0;
	}
}



static unsigned char joystick_fuzzy(int pos)
{
	return MAX(0, pos - 24);
}



static void timeloop_handler(void)
{
	if (true_time < max_true_time) {
		int i;
		int ptr = timeloop_true_ptr;
		for (i = 0; i < N_KEY_SETS; i++) {
			USER_KEYS *k = timeloop_keys[i];
			if (k) {
				volatile USER_KEYS_HELD *kh = &timeloop_keys_held[ptr][i];
				switch (k->device) {
					case UK_DEVICE_JOYSTICK:
						/* The main program polls the joystick often and
						 * saves a snapshot of the state in locked memory for
						 * us. We cannot use the joy[] array directly from
						 * here because it is not locked.
						 */
						kh->left  = joystick_fuzzy(joystick_left [i]);
						kh->right = joystick_fuzzy(joystick_right[i]);
						kh->accel = joystick_fuzzy(joystick_accel[i]);
						kh->land  = joystick_land[i] >= 48;
						kh->fire  = joystick_fire[i] >= 48;
						break;
					default: /* UK_DEVICE_KEYBOARD */
						kh->left  = key[k->left ];
						kh->right = key[k->right];
						kh->accel = key[k->accel];
						kh->land  = key[k->land ];
						kh->fire  = key[k->fire ];
				}
			}
		}
		timeloop_true_ptr = (ptr + 1) & MAX_SKIP;
		true_time++;
	} else
		n_skipped++;
	need_poll = 1;
}
END_OF_STATIC_FUNCTION(timeloop_handler);



void init_timeloop(void)
{
	LOCK_VARIABLE(true_time);
	LOCK_VARIABLE(max_true_time);
	LOCK_VARIABLE(n_skipped);
	LOCK_VARIABLE(timeloop_keys);
	LOCK_VARIABLE(timeloop_keys_held);
	LOCK_VARIABLE(timeloop_true_ptr);
	LOCK_VARIABLE(need_poll);
	LOCK_VARIABLE(joystick_left);
	LOCK_VARIABLE(joystick_right);
	LOCK_VARIABLE(joystick_fire);
	LOCK_VARIABLE(joystick_suicide);
	LOCK_VARIABLE(joystick_fuzzy);
	LOCK_FUNCTION(timeloop_handler);
}



void timeloop_add_keys(USER_KEYS *keys, int i)
{
	ASSERT(i < N_KEY_SETS);

	LOCK_DATA(keys, sizeof(*keys));
	timeloop_keys[i] = keys;
}



void timeloop_remove_keys(USER_KEYS *keys, int i)
{
	ASSERT(i < N_KEY_SETS);

	timeloop_keys[i] = NULL;
	UNLOCK_DATA(keys, sizeof(*keys));
	(void)keys; /* UNLOCK_DATA() does nothing on many platforms. */
}



/* The parameters set by this macro are designed to trick timeloop() into
 * starting afresh. By 'starting afresh', I mean that the contents of the
 * initial if-block will leave the variables in the same state as they would
 * be initialised to if timeloop_active were 0 (this includes fps_flags[]
 * containing zeros and fps_n_* being zero), and (*update)() will not be
 * called by this if-block.
 */
#define SET_NEW_TIMELOOP_PARAMETERS() \
{                                     \
	max_true_time = 0;                \
	true_time = 0;                    \
	logic_time = -1;                  \
	timeloop_logic_ptr = -1;          \
	timeloop_true_ptr = 0;            \
	need_poll = 0;                    \
	fps_n_updates = 0;                \
	fps_n_draws = 0;                  \
	n_skipped_base = n_skipped = 0;   \
	memset(fps_flags, 0, FPS_FRAMES); \
	fps_flags_current = 0;            \
}



/* Notes for timeloop():
 *
 * On call to (*update)(), logic_time corresponds to the frame that has just
 * been drawn (if there was time to draw it). The first frame is 0, and will
 * always be drawn before any updates occur.
 *
 * Recursive structures such as menus can be set up using recursive calls to
 * timeloop(). Proper protocol is to call timeloop() from the end of the
 * (*update)() function. logic_time will be incremented before the inner
 * (*update)() function is called, and the inner (*draw)() function will be
 * used immediately (with the new logic_time) if time permits. When the inner
 * (*update)() function returns 1, logic_time will again be incremented
 * before the outer (*update)() function is called, and the outer (*draw)()
 * function will be used immediately (with the new logic_time) if time
 * permits.
 */
int timeloop(
	int speed,
	int max_skip,
	void (*draw)(void *data),
	int (*update)(void *data),
	void *data
)
{
	int rv;

	ASSERT(max_skip <= MAX_SKIP);

	if (timeloop_active++) {
		logic_time++;
		timeloop_logic_ptr = (timeloop_logic_ptr + 1) & MAX_SKIP;

		FPS_FLAGS_SET(fps_flags_current);
		fps_flags_current = FPS_UPDATE_CALLED;

		while (logic_time < true_time) {
			poll_stuff();
			ckey = (/* non-volatile */ USER_KEYS_HELD *)timeloop_keys_held[timeloop_logic_ptr];
			rv = (*update)(data);

			if (rv) {
				timeloop_active--;
				return rv;
			}

			logic_time++;
			timeloop_logic_ptr = (timeloop_logic_ptr + 1) & MAX_SKIP;

			FPS_FLAGS_SET(fps_flags_current);
			fps_flags_current = FPS_UPDATE_CALLED;
		}
	} else {
		timeloop_speed = speed;

		true_time = logic_time = 0;
		max_true_time = 0;

		timeloop_logic_ptr = timeloop_true_ptr = 0;

		need_poll = 0;

		fps_n_updates = 0;
		fps_n_draws = 0;
		n_skipped_base = n_skipped = 0;
		memset(fps_flags, 0, FPS_FRAMES);
		fps_flags_current = FPS_UPDATE_CALLED;

		install_int_ex(&timeloop_handler, speed);
	}

	for (;;) {
		max_true_time = logic_time + max_skip;

		/* != instead of < to allow for wrapping */
		while (n_skipped_base != n_skipped) {
			FPS_FLAGS_SET(0);
			n_skipped_base++;
		}

		poll_stuff();
		(*draw)(data);
		fps_flags_current = FPS_UPDATE_CALLED | FPS_DRAW_CALLED;

		while (logic_time >= true_time) {
			poll_stuff();
			yield_timeslice();
		}

		do {
			poll_stuff();
			ckey = (/* non-volatile */ USER_KEYS_HELD *)timeloop_keys_held[timeloop_logic_ptr];
			rv = (*update)(data);

			if (rv) {
				if (--timeloop_active == 0)
					remove_int(timeloop_handler);
				return rv;
			}

			logic_time++;
			timeloop_logic_ptr = (timeloop_logic_ptr + 1) & MAX_SKIP;

			FPS_FLAGS_SET(fps_flags_current);
			fps_flags_current = FPS_UPDATE_CALLED;
		} while (logic_time < true_time);
	}
}



/* new_timeloop() can be used as an alternative to timeloop() when recursing.
 * For the inner functions, logic_time will count from 0. logic_time will
 * again be reset to 0 when the outer function resumes. There will be
 * discontinuities; before and after calling this function you can execute
 * time-consuming algorithms, and the computer will not try to catch up. This
 * function also allows the speed to be changed, whereas sub-timeloops will
 * normally inherit the speed of the caller.
 */
int new_timeloop(
	int speed,
	int max_skip,
	void (*draw)(void *data),
	int (*update)(void *data),
	void *data
)
{
	int rv;
	int parent_speed;

	ASSERT(timeloop_active);

	parent_speed = timeloop_speed;
	timeloop_speed = speed;
	install_int_ex(&timeloop_handler, speed);

	SET_NEW_TIMELOOP_PARAMETERS();

	rv = timeloop(speed, max_skip, draw, update, data);

	SET_NEW_TIMELOOP_PARAMETERS();

	timeloop_speed = parent_speed;
	install_int_ex(&timeloop_handler, parent_speed);

	return rv;
}
