#include <math.h>
#include "main.h"
#include "game.h"
#include "timeloop.h"
#include "collide.h"
#include "drawblt.h"
#include "explode.h"
#include "drawpwrp.h"



#define FRICTION_UNIT 24.0f

static void apply_friction(fixed *_a, fixed *_b, fixed *_c, float cscale)
{
	float a = *_a, b = *_b, c = *_c * cscale;
	float d = a*a + b*b + c*c;
	if (d < FRICTION_UNIT * FRICTION_UNIT) {
		*_a = *_b = *_c = 0;
		return;
	}
	d = 1.0f - FRICTION_UNIT / sqrt(d);
	*_a = (int)(a*d);
	*_b = (int)(b*d);
	*_c = (int)(c*d / cscale);
}



#define MAX_SPEED 0x4000

static void limit_speed(fixed *_a, fixed *_b)
{
	float a = *_a, b = *_b;
	float d = a*a + b*b;
	if (d > (float)MAX_SPEED*MAX_SPEED) {
		d = MAX_SPEED / sqrt(d);
		*_a = (int)(a*d);
		*_b = (int)(b*d);
	}
}



#define BULLET_RECOIL_SHIFT 7

static void fire_bullet(GAME *game, POD *pod)
{
	BULLET *bullet = malloc(sizeof(*bullet));
	if (!bullet) return;
	bullet->next = game->bullet;
	game->bullet = bullet;
	{
		float cp = (float)cos(pod->phi * (3.14159265358979323846f / (128 << 16)));
		float sp = (float)sin(pod->phi * (3.14159265358979323846f / (128 << 16)));
		bullet->x = pod->x + (int)(sp * FIXED_POD_RADIUS);
		bullet->y = pod->y - (int)(cp * FIXED_POD_RADIUS);
		bullet->z = pod->z + FIXED_POD_HEIGHT;
		bullet->xv = pod->xv + (int)(sp * 0x4000);
		bullet->yv = pod->yv - (int)(cp * 0x4000);
		bullet->zv = 0;
		pod->xv -= (int)(sp * (0x4000 >> BULLET_RECOIL_SHIFT));
		pod->yv += (int)(cp * (0x4000 >> BULLET_RECOIL_SHIFT));
		limit_speed(&pod->xv, &pod->yv);
	}
	init_bullet_polygons(bullet);
}



static void deduct_health(POD *pod, int penalty)
{
	if (pod->invinc_time == 0) {
		pod->health -= penalty;
		if (pod->health <= 0) pod->health = 0;
	}
}



static void give_health(POD *pod, int gift)
{
	pod->health += gift;
	if (pod->health > 2*POD_HEALTH_INITIAL) pod->health = 2*POD_HEALTH_INITIAL;
}



static int random_empty_tile(GAME *game, int *_x, int *_y)
{
	int n = 0;
	int x, y;
	for (y = 0; y < game->map_h; y++) {
		for (x = 0; x < game->map_w; x++) {
			if (!TILE_IS_HIGH(game->map[y][x])) {
				if (rand() % ++n == 0) {
					*_x = x;
					*_y = y;
				}
			}
		}
	}
	return n == 0;
}



static void create_powerup(GAME *game)
{
	int x, y;
	POWERUP *powerup;
	if (random_empty_tile(game, &x, &y)) return;
	powerup = malloc(sizeof(*powerup));
	if (!powerup) return;
	powerup->next = game->powerup;
	game->powerup = powerup;
	powerup->x = x;
	powerup->y = y;
	powerup->frame = rand() & (POWERUP_N_FRAMES - 1);
	powerup->type = rand() % POWERUP_N;
	init_powerup_polygons(powerup);
	game->n_powerups++;
}



static void update_pods(GAME *game)
{
	{
		POD *pod;
		for (pod = game->pod; pod; pod = pod->next) {
			if (pod->state != POD_STATE_DEAD) {
				if (pod->health <= 0) {
					create_explosion(game, fixtof(pod->x), fixtof(pod->y), fixtof(pod->z + FIXED_POD_RADIUS/3), POD_RADIUS * 2.0f);
					pod->state = POD_STATE_DEAD;
				} else {
					if (pod->invinc_time) pod->invinc_time--;
					if (pod->cloak_time) pod->cloak_time--;
				}
			}

			switch (pod->state) {
				case POD_STATE_FALLING:
				case POD_STATE_THRUSTING:
					if (ckey[pod->id].land || pod->z >= FIXED_HOVER_MAX)
						pod->state = POD_STATE_FALLING;
					else if (pod->z < FIXED_HOVER_MIN)
						pod->state = POD_STATE_THRUSTING;

					if (pod->state == POD_STATE_THRUSTING)
						pod->zv += FIXED_GRAVITY;
					else
						pod->zv -= FIXED_GRAVITY;

					pod->zv -= pod->zv >> 6;

					if (pod->z > 0) {
						if (ckey[pod->id].left && !ckey[pod->id].right) {
							pod->phiv -= 0x300;
							if (pod->phiv < -0x100000) pod->phiv = -0x100000;
						} else if (ckey[pod->id].right && !ckey[pod->id].left) {
							pod->phiv += 0x300;
							if (pod->phiv > 0x100000) pod->phiv = 0x100000;
						}
						if (ckey[pod->id].accel) {
							pod->xv += (int)(0x18 * sin(pod->phi * (3.14159265358979323846f / (128 << 16))));
							pod->yv -= (int)(0x18 * cos(pod->phi * (3.14159265358979323846f / (128 << 16))));
							limit_speed(&pod->xv, &pod->yv);
						}
					}

					pod->x += pod->xv;
					pod->y += pod->yv;

					detect_wall_collisions(game, pod);

					pod->z += pod->zv;
					pod->phi = (pod->phi + pod->phiv) & 0xFFFFFF;

					if (pod->z <= 0) {
						pod->z = 0;
						pod->zv = 0;
						apply_friction(&pod->xv, &pod->yv, &pod->phiv, 1.0f / 32.0f);
					}

					if (pod->fire_delay > 0)
						pod->fire_delay--;
					else if (ckey[pod->id].fire) {
						fire_bullet(game, pod);
						pod->fire_delay = POD_FIRE_DELAY;
					}

					break;

				default: /* POD_STATE_DEAD */
					break;
			}
		}
	}

	{
		POD *pod1, *pod2;
		for (pod2 = game->pod; pod2; pod2 = pod2->next)
			if (POD_IS_THERE(pod2))
				for (pod1 = pod2->next; pod1; pod1 = pod1->next)
					if (POD_IS_THERE(pod1))
						detect_pod_collision(game, pod1, pod2);
	}

	{
		BULLET **bulletp = &game->bullet;
		while (*bulletp) {
			BULLET *bullet = *bulletp;
			int xi, yi;
			bullet->zv -= FIXED_GRAVITY;
			bullet->x += bullet->xv;
			bullet->y += bullet->yv;
			bullet->z += bullet->zv;
			xi = bullet->x >> 16;
			yi = bullet->y >> 16;
			if (bullet->z < 0 || xi < 0 || xi >= game->map_w || yi < 0 || yi >= game->map_h || (bullet->z < FIXED_WALL_HEIGHT && TILE_IS_HIGH(game->map[yi][xi]))) {
				bullet->x -= bullet->xv >> 1;
				bullet->y -= bullet->yv >> 1;
				bullet->z -= bullet->zv >> 1;
				create_explosion(game, fixtof(bullet->x), fixtof(bullet->y), fixtof(bullet->z), 0.2f);
				*bulletp = bullet->next;
				free(bullet);
			} else {
				POD *pod;
				for (pod = game->pod; pod; pod = pod->next) {
					if (POD_IS_THERE(pod) && bullet->z >= pod->z && bullet->z < pod->z + FIXED_POD_FULL_HEIGHT) {
						float xd = (float)(pod->x - bullet->x);
						float yd = (float)(pod->y - bullet->y);
						float d = xd*xd + yd*yd;
						if (d < (float)FIXED_POD_RADIUS*FIXED_POD_RADIUS) {
							pod->xv += bullet->xv >> BULLET_RECOIL_SHIFT;
							pod->yv += bullet->yv >> BULLET_RECOIL_SHIFT;
							limit_speed(&pod->xv, &pod->yv);
							deduct_health(pod, 0x800);
							create_explosion(game, fixtof(bullet->x), fixtof(bullet->y), fixtof(bullet->z), 0.3f);
							*bulletp = bullet->next;
							free(bullet);
							goto skip;
						}
					}
				}
				bulletp = &bullet->next;
			}
			skip: ;
		}
	}

	if (game->new_powerup_time == 0) {
		if (game->n_powerups < MAX_POWERUPS) {
			create_powerup(game);
			game->new_powerup_time = NEW_POWERUP_TIME - 1;
		}
	} else
		game->new_powerup_time--;

	{
		POWERUP **powerupp = &game->powerup;
		while (*powerupp) {
			POWERUP *powerup = *powerupp;
			POD *pod;
			fixed x = (powerup->x << 16) + (1 << 15);
			fixed y = (powerup->y << 16) + (1 << 15);
			powerup->frame = (powerup->frame + 1) & (POWERUP_N_FRAMES - 1);
			for (pod = game->pod; pod; pod = pod->next) {
				if (POD_IS_THERE(pod)) {
					float xd = x - pod->x, yd = y - pod->y;
					if (xd*xd + yd*yd < (float)(FIXED_POWERUP_RADIUS+FIXED_POD_RADIUS)*(FIXED_POWERUP_RADIUS+FIXED_POD_RADIUS)) {
						switch (powerup->type) {
							case POWERUP_INVINCIBILITY:
								pod->invinc_time = POD_INVINC_TIME;
								break;
							case POWERUP_CLOAK:
								pod->cloak_time = POD_CLOAK_TIME;
								break;
							default: /* POWERUP_EXTRA_HEALTH */
								give_health(pod, 0x2000);
								break;
						}
						*powerupp = powerup->next;
						free(powerup);
						game->n_powerups--;
						goto skip2;
					}
				}
			}
			powerupp = &powerup->next;
			skip2: ;
		}
	}

	update_explosions(game);
}



int update_game(void *data)
{
	GAME *game = data;

	update_pods(game);

	while (keypressed()) {
		int k = readkey();
		if (k >> 8 == KEY_ESC) return 1;
	}

	return 0;
}
